//***************************************************************************
// Ces applets viennent en complment du livre :
//
//  		VIE ARTIFICIELLE
//      O la biologie rencontre l'informatique
//		Illustr en Java
//            Jean-Philippe Rennard
//           	  Vuibert 2002
//===========================================================================
//  Algorithme Gntique : Problme du Voyageur de Commerce  (Vuibert)
//    v:0.1 10/09/2002
//
//    This program is free software; you can redistribute it and/or
//    modify it under the terms of the GNU General Public License
//    as published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//***************************************************************************
package pvcag;

/** Classe de gestion d'un individu */
public class Cindividu {

  /** Gnome = parcours */
  private String genome;
  /** Distance parcourue */
  private long fitnessBrute;
  /** Fitness */
  private double fitness;

  /** Constructeur par dfaut : individu alatoire */
  public Cindividu(int nbvilles) {
    fitness=0;
    fitnessBrute=0;
    genome="";
    for(byte i=1;i<nbvilles;i++)
      genome += (char)(i);
    String g="";
    g += (char)(0);
    // Permutations de Josephus.
    // sur une ide de Scott Robert Ladd (http://www.coyotegulch.com).
    boolean [] select=new boolean[nbvilles];
    int c=(int)(Math.random()*nbvilles-1);
    int l=(int)(Math.random()*(nbvilles-3)+1);
    int rang=0;
    while(true) {
      g += genome.charAt(c);
      select[c]=true;
      if(rang==nbvilles-2)
        break;
      for(int i=0;i<l;i++) {
        do {
          c++;
          if(c>nbvilles-2) c=0;
        } while(select[c]==true);
      }
      rang++;
    }
    g += (char)(0);
    genome=g;
  }

  /** Constructeur d'un descendant */
  public Cindividu(Cindividu pere, Cindividu mere, int typecross) {
    genome="";
    switch(typecross) {
      case Cag.CPA:
        genome=cpa(pere,mere);
      break;
      case Cag.INVERSION:
        genome=inversion(pere.genome);
      break;
      default:
        genome=pere.genome;
    }
    fitnessBrute=0;
    fitness=0;
  }
  /** Interversion */
  private String interversion(String s, int d, int f) {
    if(d==f)
      return(s);
    int deb=(d<f ? d : f);
    int fin=(deb==d ? f : d);
    String t="";
    t +=  s.substring(0,deb);
    t += s.charAt(fin);
    t += s.substring(deb+1,fin);
    t += s.charAt(deb);
    t += s.substring(fin+1, s.length());
    return t;
  }
  /** Inversion */
  private String inversion(String s) {
    int cross1=(int)Math.floor(Math.random()*(s.length()-3))+1;
    int cross2=cross1+
          (int)Math.floor(Math.random()*(s.length()-cross1-2)+1);
    String t="";
    t += s.substring(0,cross1);
    for(int i=cross2;i>=cross1;i--)
      t += s.charAt(i);
    t += s.substring(cross2+1);
    return t;
  }
  /** Crossover partiellement assortie */
  private String cpa(Cindividu pere,Cindividu mere) {
    int cross1=(int)Math.floor(Math.random()*(pere.genome.length()-3))+1;
    int cross2=cross1+
          (int)Math.floor(Math.random()*(pere.genome.length()-cross1-2)+1);
    Cindividu base, second;
    base=(Math.random()<0.5 ? pere : mere);
    second=(base==pere ? mere : pere);
    // chaine de croisement
    String cross=second.genome.substring(cross1, cross2+1);
    String b=base.genome;
    for(int i=cross1;i<cross2+1;i++)
        b=interversion(b,i,b.indexOf(second.genome.charAt(i)));
    return b;
  }
  /** Mutations (interversion) */
  public void mute(double prob) {
    if(Math.random()>prob)
      return;
    int pos1=(int)Math.floor(Math.random()*(genome.length()-2)+1);
    int pos2;
    do {
      pos2=(int)Math.floor(Math.random()*(genome.length()-2)+1);
    } while (pos1==pos2);
    genome=interversion(genome, pos1, pos2);
  }
  /** Modificateur de la valeur de fitness brute */
  public void setFitnessBrute(long f) {
    fitnessBrute=f;
  }
  /** Accesseur de la valeur de fitness brute */
  public long getFitnessBrute() {
    return fitnessBrute;
  }
  /** Modificateur de la valeur de fitness */
  public void setFitness(double f) {
    fitness=f;
  }
  /** Accesseur de la valeur de fitness */
  public double getFitness() {
    return fitness;
  }
  /** Accesseur du gnome */
  public String getGenome() {
    return genome;
  }
}